; zero page variables for the ANSI terminal parser
TERMINAL_PARSE_STATE=$f3
TERMINAL_PARSE_V_COUNT=$f4
TERMINAL_PARSE_V=$f5 ; - $f9
TERMINAL_PARSE_COLOUR=$f9
TERMINAL_TOP_LINE=$fa   ; first visible line on display (0-index)

; $fb-$ff is used to store intermediate calculations during vsync
TERMINAL_XX=$fb
TERMINAL_XX_lo=$fb
TERMINAL_XX_hi=$fc
TERMINAL_YY=$fd
TERMINAL_YY_lo=$fd
TERMINAL_YY_hi=$fe

TERMINAL_STATE_GND=0
TERMINAL_STATE_ESC=1
TERMINAL_STATE_CSI=2

; sets up sprites #1,#2 to show the cursor. should be called during vsync
; interrupt handler. does not interfere with vera_addr storing cursor position.
terminal_vera_show_cursor:
    lda vera_addr_mid   ; line of cursor
    sec
    sbc terminal_top_line ; get line number relative to display
    sta terminal_yy_lo
    stz terminal_yy_hi
    ; multiply by 16
    .rept 4
        asl terminal_yy_lo
        rol terminal_yy_hi
    .next
    lda vera_addr_lo ; column * 2
    sta terminal_xx_lo
    stz terminal_xx_hi
    ; multiply by 4 to get column*8
    .rept 2
        asl terminal_xx_lo
        rol terminal_xx_hi
    .next
    ; set up vera addressing
    lda #1
    sta vera_ctrl       ; address port 1
    lda #$11
    sta vera_addr_hi
    lda #$fc
    sta vera_addr_mid
    lda #$08
    sta vera_addr_lo
    stz vera_ctrl       ; return to port 0

    ; write sprite 1 data
    stz vera_data1
    stz vera_data1 ; bitmap address
    lda terminal_xx_lo
    sta vera_data1
    lda terminal_xx_hi
    sta vera_data1
    lda terminal_yy_lo
    sta vera_data1
    lda terminal_yy_hi
    sta vera_data1
    lda #$0c
    sta vera_data1  ; enable
    lda #$41    ; width/height , palette
    sta vera_data1

    ; write sprite 2 data
    ; TODO second sprite should show the character under the cursor in
    ; reverse video

    rts

; set the horizontal position of the cursor to the value in .A
; first column is #1. zero is interpreted as one.
terminal_cursor_hor:
    ; decrement .A unless it is already zero.
    ; this converts 1-index to 0-index, but ensures that a value of 0 is
    ; interpreted as a 1.
    cmp #0
    beq +
        dea
+   ; make sure that this value falls within bounds
    cmp #0
    bpl +
        ; if position is negative, reset to zero
        lda #0
+   cmp #80
    bcc +
        ; if position is >=80, return to 79
        lda #79
+
    ; position is within bounds, set the cursor position
    asl ; because each character takes two bytes
    sta vera_addr_lo
    rts

; set the vertical position of the cursor to the value in .A
; first visible row is #1. zero is interpreted as one.
terminal_cursor_vert:
    ; decrement .A unless it is already zero
    cmp #0
    beq +
        dea
+   ; check bounds
    ; because there is a separate routine for relative motion, this routine
    ; does not need to worry about negative input. input value is treated as
    ; unsigned and values larger than 29 are saturated
    cmp #30
    bcc +
        lda #29
        clc
+   ; add to the top visible line number
    adc terminal_top_line
    ; move cursor
    sta vera_addr_mid
    rts

; adjusts the vertical position (downward) of the cursor by the value in .A
; ( ) -- ( )
terminal_cursor_vert_rel:
    clc
    adc vera_addr_mid       ; .A is new current line
    and #$7f                ; take 7 bits for 128 lines
    sta vera_addr_mid       ; .A is new line
    ; TODO check scroll up
    sec
    sbc terminal_top_line
    ; this subtraction will result in a negative at the seem between the last
    ; line and the first line looping back. this is corrected by adding 128
    ; (which is equal to 0 mod the number of lines)
    bcs +
        ; carry is clear
        adc #$80    ; carry is set after addition
+
    sbc #29
    bcc _no_scroll
        ; scroll down by the amount in .A
        jmp terminal_scroll_down
_no_scroll:
    rts

; scrolls the display down the number of lines in .A
; TODO this ought to clear new lines as they come in
terminal_scroll_down:
    clc
    adc terminal_top_line
    ; take 7 bits (because there are 128 lines)
    and #$7f
    sta terminal_top_line
    rts

; input in .A
; ( ) -- ( )
terminal_insertch:
    ; select routine based on TERMINAL_PARSE_STATE
    ldy terminal_parse_state
    cpy #terminal_state_gnd
    bne +
        jmp terminal_parse_gnd
+   cpy #terminal_state_esc
    bne +
        jmp terminal_parse_esc
+   cpy #terminal_state_csi
        jmp terminal_parse_csi
    ; invalid state ?!
    rts

; default terminal state. printable characters are sent to output.
terminal_parse_gnd:
    cmp #$1b                    ; escape
    bne _not_escape
    ; transition state to ESC and exit
    lda #terminal_state_esc
    sta terminal_parse_state
    rts
_not_escape
    ; process input byte as utf-8
    jsr utf8_parse_byte
    beq _output
    rts
_output:
    ; output this character to screen
    lda utf8_parse_codept+1
    d_push
    lda utf8_parse_codept
    d_push                      ; ( Codept )
    jmp terminal_printch

; last byte input was escape. this will proceed to CSI if next character is '[',
; otherwise return to ground. no characters are output.
terminal_parse_esc:
    cmp #'['
    bne _not_csi
    ; go to csi state
    lda #terminal_state_csi
    sta terminal_parse_state
    ; set initial values to read parameters
    stz terminal_parse_v_count  ; there is one parameter
    stz terminal_parse_v        ; and it's zero
    rts
_not_csi:
    ; return to ground
    stz terminal_parse_state
    rts

; previous characters were ^[[, starting a control sequence. possibilities:
; - $20-$2f intermediate bytes (currently ignored)
; - $30-$3f parameter bytes
; - $40-$7f final byte (return to ground)
terminal_parse_csi:
    tay                             ; .Y is backup of byte in
    and #$f0
    cmp #$20
    bne _not_intermediate
        ; intermediate bytes (ignore them)
        rts
_not_intermediate:
    cmp #$30
    bne _not_parameter
        ; parameter byte
        tya
        jmp terminal_parse_csi_parameter
_not_parameter:
    bit #$80
    bne _not_final
        ; final byte
        ; return to ground state and execute the control string
        ; (if it's supported)
        stz terminal_parse_state
        tya
        jmp terminal_parse_csi_final
_not_final:
    ; invalid byte--return to ground and do nothing
    stz terminal_parse_state
    rts

; read a parameter byte from .A
terminal_parse_csi_parameter:
    cmp #';'
    bne _not_semicolon
    ; add another parameter
    inc terminal_parse_v_count
    ldy terminal_parse_v_count
    lda #0
    sta terminal_parse_v,y  ; zero the parameter
    rts
_not_semicolon:
    ; I should probably be checking that this is *actually* a digit.
    ; oh well, if it isn't it's part of a CSI string that is not supported
    and #$0f
    d_push                              ; ( next_digit )
    ; multiply the current value by ten and add next_digit
    ldy terminal_parse_v_count
    lda terminal_parse_v,y  ; current parameter = p
    asl                                 ; .A = 2p
    d_push                              ; ( next_digit 2p )
    asl
    asl                                 ; .A = 8p
    clc
    adc 0,x                             ; .A = 8p+2p
    adc 1,x                             ; .A = 10p + next_digit
    ; save intermediate parameter
    sta terminal_parse_v,y
    d_drop
    d_drop                              ; ( )
    rts

; parse final byte of csi
terminal_parse_csi_final:
    ; convert ascii to index into table
    sec
    sbc #$40
    asl
    sta r0_lo
    ; high byte from jump table
    lda #>terminal_csi_dispatch
    sta r0_hi
    ; dereference address
    lda (r0)
    sta r1_lo
    ldy #1
    lda (r0),y
    sta r1_hi
    jmp (r1)

; writes given unicode character to screen
; ( Codept ) -- ( )
terminal_printch:
    lda 1,x                                 ; Codept hi
    bne _output ; control characters are on unicode page 0
    lda 0,x                                 ; Codept lo
    cmp #$07
    bne _not_bel
        ; TODO make some kind of bell noise
        bra _return
_not_bel:
    cmp #$08    ; back space
    bne _not_bs
        lda vera_addr_lo
        beq _return ; skip if cursor is already at beginning of a line
        sec
        sbc #2
        sta vera_addr_lo
        bra _return
_not_bs:
    cmp #$0a    ; line feed
    bne  _not_lf
        lda #1
        jsr terminal_cursor_vert_rel ; move cursor down
        bra _return
_not_lf:
    cmp #$0d    ; carriage return
    bne _not_cr
        stz vera_addr_lo
        bra _return
_not_cr:
_output:
    jsr unicode_lookup_screen_code          ; ( Scrcode )
    lda 0,x                                 ; Scrcode lo
    sta vera_data0
    lda 1,x                                 ; Scrcode hi
    eor terminal_parse_colour
    sta vera_data0
    ; check if the line is overflowed
    lda vera_addr_lo
    cmp #81*2
    bne _return
    stz vera_addr_lo
    inc vera_addr_mid
_return:
    d_drop
    d_drop                                  ; ( )
    rts
